@@ -27,7 +27,10 @@ package org.sqlite;
 import java.io.*;
 import java.net.URL;
 import java.net.URLConnection;
-import java.nio.file.*;
+import java.nio.file.Files;
+import java.nio.file.Path;
+import java.nio.file.Paths;
+import java.nio.file.StandardCopyOption;
 import java.security.DigestInputStream;
 import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
@@ -36,12 +39,15 @@ import java.util.List;
 import java.util.Properties;
 import java.util.UUID;
 import java.util.stream.Stream;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
+import org.sqlite.util.LibraryLoaderUtil;
 import org.sqlite.util.OSInfo;
 import org.sqlite.util.StringUtils;
 
 /**
  * Set the system properties, org.sqlite.lib.path, org.sqlite.lib.name, appropriately so that the
- * SQLite JDBC driver can find *.dll, *.jnilib and *.so files, according to the current OS (win,
+ * SQLite JDBC driver can find *.dll, *.dylib and *.so files, according to the current OS (win,
  * linux, mac).
  *
  * <p>The library files are automatically extracted from this project's package (JAR).
@@ -51,6 +57,7 @@ import org.sqlite.util.StringUtils;
  * @author leo
  */
 public class SQLiteJDBCLoader {
+    private static final Logger logger = LoggerFactory.getLogger(SQLiteJDBCLoader.class);
 
     private static final String LOCK_EXT = ".lck";
     private static boolean extracted = false;
@@ -94,37 +101,15 @@ public class SQLiteJDBCLoader {
                                     try {
                                         Files.delete(nativeLib);
                                     } catch (Exception e) {
-                                        System.err.println(
-                                                "Failed to delete old native lib: "
-                                                        + e.getMessage());
+                                        logger.error("Failed to delete old native lib", e);
                                     }
                                 }
                             });
         } catch (IOException e) {
-            System.err.println("Failed to open directory: " + e.getMessage());
+            logger.error("Failed to open directory", e);
         }
     }
 
-    /**
-     * @return True if the SQLite JDBC driver is set to pure Java mode; false otherwise.
-     * @deprecated Pure Java no longer supported
-     */
-    @Deprecated
-    static boolean getPureJavaFlag() {
-        return Boolean.parseBoolean(System.getProperty("sqlite.purejava", "false"));
-    }
-
-    /**
-     * Checks if the SQLite JDBC driver is set to pure Java mode.
-     *
-     * @return True if the SQLite JDBC driver is set to pure Java mode; false otherwise.
-     * @deprecated Pure Java nolonger supported
-     */
-    @Deprecated
-    public static boolean isPureJavaMode() {
-        return false;
-    }
-
     /**
      * Checks if the SQLite JDBC driver is set to native mode.
      *
@@ -191,7 +176,8 @@ public class SQLiteJDBCLoader {
      * @return
      */
     private static boolean extractAndLoadLibraryFile(
-            String libFolderForCurrentOS, String libraryFileName, String targetFolder) {
+            String libFolderForCurrentOS, String libraryFileName, String targetFolder)
+            throws FileException {
         String nativeLibraryFilePath = libFolderForCurrentOS + "/" + libraryFileName;
         // Include architecture name in temporary filename in order to avoid conflicts
         // when multiple JVMs with different architectures running at the same time
@@ -227,7 +213,7 @@ public class SQLiteJDBCLoader {
                 try (InputStream nativeIn = getResourceAsStream(nativeLibraryFilePath);
                         InputStream extractedLibIn = Files.newInputStream(extractedLibFile)) {
                     if (!contentsEquals(nativeIn, extractedLibIn)) {
-                        throw new RuntimeException(
+                        throw new FileException(
                                 String.format(
                                         "Failed to write a native library file at %s",
                                         extractedLibFile));
@@ -236,7 +222,7 @@ public class SQLiteJDBCLoader {
             }
             return loadNativeLibrary(targetFolder, extractedLibFileName);
         } catch (IOException e) {
-            e.printStackTrace();
+            logger.error("Unexpected IOException", e);
             return false;
         }
     }
@@ -260,7 +246,7 @@ public class SQLiteJDBCLoader {
             connection.setUseCaches(false);
             return connection.getInputStream();
         } catch (IOException e) {
-            e.printStackTrace();
+            logger.error("Could not connect", e);
             return null;
         }
     }
@@ -280,12 +266,12 @@ public class SQLiteJDBCLoader {
                 System.load(new File(path, name).getAbsolutePath());
                 return true;
             } catch (UnsatisfiedLinkError e) {
-                System.err.println(
-                        "Failed to load native library:"
-                                + name
-                                + ". osinfo: "
-                                + OSInfo.getNativeLibFolderPathForCurrentOS());
-                e.printStackTrace();
+
+                logger.error(
+                        "Failed to load native library: {}. osinfo: {}",
+                        name,
+                        OSInfo.getNativeLibFolderPathForCurrentOS(),
+                        e);
                 return false;
             }
 
@@ -294,6 +280,16 @@ public class SQLiteJDBCLoader {
         }
     }
 
+    private static boolean loadNativeLibraryJdk() {
+        try {
+            System.loadLibrary(LibraryLoaderUtil.NATIVE_LIB_BASE_NAME);
+            return true;
+        } catch (UnsatisfiedLinkError e) {
+            logger.error("Failed to load native library through System.loadLibrary", e);
+            return false;
+        }
+    }
+
     /**
      * Loads SQLite native library using given path and name of the library.
      *
@@ -310,10 +306,7 @@ public class SQLiteJDBCLoader {
         String sqliteNativeLibraryPath = System.getProperty("org.sqlite.lib.path");
         String sqliteNativeLibraryName = System.getProperty("org.sqlite.lib.name");
         if (sqliteNativeLibraryName == null) {
-            sqliteNativeLibraryName = System.mapLibraryName("sqlitejdbc");
-            if (sqliteNativeLibraryName != null && sqliteNativeLibraryName.endsWith(".dylib")) {
-                sqliteNativeLibraryName = sqliteNativeLibraryName.replace(".dylib", ".jnilib");
-            }
+            sqliteNativeLibraryName = LibraryLoaderUtil.getNativeLibName();
         }
 
         if (sqliteNativeLibraryPath != null) {
@@ -326,22 +319,9 @@ public class SQLiteJDBCLoader {
         }
 
         // Load the os-dependent library from the jar file
-        String packagePath = SQLiteJDBCLoader.class.getPackage().getName().replaceAll("\\.", "/");
-        sqliteNativeLibraryPath =
-                String.format(
-                        "/%s/native/%s", packagePath, OSInfo.getNativeLibFolderPathForCurrentOS());
-        boolean hasNativeLib = hasResource(sqliteNativeLibraryPath + "/" + sqliteNativeLibraryName);
-
-        if (!hasNativeLib) {
-            if (OSInfo.getOSName().equals("Mac")) {
-                // Fix for openjdk7 for Mac
-                String altName = "libsqlitejdbc.jnilib";
-                if (hasResource(sqliteNativeLibraryPath + "/" + altName)) {
-                    sqliteNativeLibraryName = altName;
-                    hasNativeLib = true;
-                }
-            }
-        }
+        sqliteNativeLibraryPath = LibraryLoaderUtil.getNativeLibResourcePath();
+        boolean hasNativeLib =
+                LibraryLoaderUtil.hasNativeLib(sqliteNativeLibraryPath, sqliteNativeLibraryName);
 
         if (hasNativeLib) {
             // temporary library folder
@@ -370,8 +350,14 @@ public class SQLiteJDBCLoader {
             }
         }
 
+        // As an ultimate last resort, try loading through System.loadLibrary
+        if (loadNativeLibraryJdk()) {
+            extracted = true;
+            return;
+        }
+
         extracted = false;
-        throw new Exception(
+        throw new NativeLibraryNotFoundException(
                 String.format(
                         "No native library found for os.name=%s, os.arch=%s, paths=[%s]",
                         OSInfo.getOSName(),
@@ -379,10 +365,6 @@ public class SQLiteJDBCLoader {
                         StringUtils.join(triedPaths, File.pathSeparator)));
     }
 
-    private static boolean hasResource(String path) {
-        return SQLiteJDBCLoader.class.getResource(path) != null;
-    }
-
     @SuppressWarnings("unused")
     private static void getNativeLibraryFolderForTheCurrentOS() {
         String osName = OSInfo.getOSName();
@@ -403,27 +385,42 @@ public class SQLiteJDBCLoader {
 
     /** @return The version of the SQLite JDBC driver. */
     public static String getVersion() {
+        return VersionHolder.VERSION;
+    }
 
-        URL versionFile =
-                SQLiteJDBCLoader.class.getResource(
-                        "/META-INF/maven/org.xerial/sqlite-jdbc/pom.properties");
-        if (versionFile == null) {
-            versionFile =
-                    SQLiteJDBCLoader.class.getResource(
-                            "/META-INF/maven/org.xerial/sqlite-jdbc/VERSION");
-        }
+    /**
+     * This class will load the version from resources during <clinit>. By initializing this at
+     * build-time in native-image, the resources do not need to be included in the native
+     * executable, and we're eliminating the IO operations as well.
+     */
+    public static final class VersionHolder {
+        private static final String VERSION;
 
-        String version = "unknown";
-        try {
-            if (versionFile != null) {
-                Properties versionData = new Properties();
-                versionData.load(versionFile.openStream());
-                version = versionData.getProperty("version", version);
-                version = version.trim().replaceAll("[^0-9\\.]", "");
+        static {
+            URL versionFile =
+                    VersionHolder.class.getResource(
+                            "/META-INF/maven/org.xerial/sqlite-jdbc/pom.properties");
+            if (versionFile == null) {
+                versionFile =
+                        VersionHolder.class.getResource(
+                                "/META-INF/maven/org.xerial/sqlite-jdbc/VERSION");
             }
-        } catch (IOException e) {
-            e.printStackTrace();
+
+            String version = "unknown";
+            try {
+                if (versionFile != null) {
+                    Properties versionData = new Properties();
+                    versionData.load(versionFile.openStream());
+                    version = versionData.getProperty("version", version);
+                    version = version.trim().replaceAll("[^0-9\\.]", "");
+                }
+            } catch (IOException e) {
+                // inline creation of logger to avoid build-time initialization of the logging
+                // framework in native-image
+                LoggerFactory.getLogger(VersionHolder.class)
+                        .error("Could not read version from file: {}", versionFile, e);
+            }
+            VERSION = version;
         }
-        return version;
     }
 }
